/***************************************************************************
 *
 * udrv.c : user space driver for APC finger print reader
 *
 * Copyright (C) 2005-2006 David Zeuthen, <david@fubar.dk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 **************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <usb.h>

static void 
hexdump (unsigned char *buf, size_t size)
{
	unsigned int i;
	unsigned int j;
	unsigned int n;

	n = 0;
	printf ("Dumping %d=0x%x bytes\n", size, size);
	while (n < size) {

		printf ("0x%04x: ", n);

		j = n;
		for (i = 0; i < 16; i++) {
			if (j >= size)
				break;
			printf ("%02x ", buf[j]);
			j++;
		}
		
		for ( ; i < 16; i++) {
			printf ("   ");
		}
		
		printf ("   ");
		
		j = n;
		for (i = 0; i < 16; i++) {
			if (j >= size)
				break;
			printf ("%c", isprint(buf[j]) ? buf[j] : '.');
			j++;
		}

		printf ("\n");
		
		n += 16;
	}
}

static void 
dump_image (unsigned char *buf, unsigned int len, int serial)
{
	int x, y;
	int i, j, k, n;
	int width;
	int height;
	FILE *f;
	char name[256];
	unsigned char *pixmap;
	int row;
	int col;

	width = 128;
	height = 128;

	pixmap = NULL;

	pixmap = calloc (sizeof (unsigned char), 3 * 128 * 128);
	if (pixmap == NULL)
		goto out;

#define PUT_PIXEL(xpos, ypos, red, green, blue) do {if ((xpos) >= 0 && (xpos) < width && (ypos) >= 0 && (ypos) < height) {pixmap [3 * ((xpos) + (height - 1 - (ypos)) * width) + 0] = (blue); pixmap [3 * ((xpos) + (height - 1 - (ypos)) * width) + 1] = (green); pixmap [3 * ((xpos) + (height - 1 - (ypos)) * width) + 2] = (red);} else printf ("Warning (%d, %d) is out of range\n", (xpos), (ypos));} while (0)


/*
	for (y = 0; y < height; y++) {
		for (x = 0; x < width; x++) {
			PUT_PIXEL (x, y, serial*16, x, y);
		}
	}
*/

	if (buf[0] == 0xe0) {
		n = 1;
		for (row = 0; row < 8; row++) {
			for (col = 0; col < 128; col++) {
				int val;
				int pixel1;
				int pixel2;
				
				for (k = 0; k < 8; k++) {
					val = buf[n++];
					pixel1 = (val & 0x0f);
					pixel2 = (val & 0xf0) >> 4;
					pixel1 = (pixel1<<5) | (pixel1<<1);
					pixel2 = (pixel2<<5) | (pixel2<<1);
					pixel1 = 0xff - pixel1;
					pixel2 = 0xff - pixel2;
					PUT_PIXEL (col, (7-row)*16 + (7-k)*2 + 1, pixel1, pixel1, pixel1);
					PUT_PIXEL (col, (7-row)*16 + (7-k)*2 + 0, pixel2, pixel2, pixel2);
				}
			}
			n++;
		}
	} else if (buf[0] == 0xc0) {
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
				PUT_PIXEL (x, y, serial*16, x, y);
			}
		}		
	}

	snprintf (name, sizeof (name), "fingerprint-%03d.tga", serial);
	if ((f = fopen (name, "wb")) == NULL)
		goto out;

	putc (0, f);
	putc (0, f);
	putc (2, f);                         /* uncompressed RGB */
	putc (0, f); 
	putc (0, f);
	putc (0, f); 
	putc (0, f);
	putc (0, f);
	putc (0, f); 
	putc (0, f);           /* X origin */
	putc (0, f); 
	putc (0, f);           /* y origin */
	putc ((width & 0x00FF), f);
	putc ((width & 0xFF00) / 256, f);
	putc ((height & 0x00FF), f);
	putc ((height & 0xFF00) / 256, f);
	putc (24,f);                        /* 24 bit bitmap */
	putc (0,f);
	fwrite (pixmap, width * height * 3, sizeof (unsigned char), f);		

	fclose (f);
out:
	if (pixmap != NULL)
		free (pixmap);
}

static void
read_from_sensor (struct usb_device *dev)
{
	int ret;
	usb_dev_handle *h;
	unsigned int num_packets;

	if ((h = usb_open (dev)) == NULL) {
		printf ("error opening device\n");
		goto out;
	}

	if ((ret = usb_set_configuration (h, 1)) < 0) {
		printf ("error setting configuration: %s\n", strerror (-ret));
		goto out;
	}

	if ((ret = usb_claim_interface (h, 0)) < 0) {
		printf ("error claiming interface: %s\n", strerror (-ret));
		goto out;
	}
#if 0
	int num_bytes_written;
	static unsigned char wbuf1[8] = {0x09, 0x02, 0x20, 0x00, 0x01, 0x01, 0x00, 0x80, 0x32};

	num_bytes_written = usb_bulk_write (h, 0x02, wbuf1, 9, 5000);
	if (num_bytes_written != 0)
		printf ("error doing write: %s (%d)\n", strerror (-num_bytes_written), -num_bytes_written);
	else
		printf ("num_bytes_written = %d\n", num_bytes_written);
	
	exit (1);
#endif


		{
			static char wbuf[] = {0x80, 0x01, 0x81, 0x07, 0x97, 0x08};
			usb_bulk_write (h, 0x02, wbuf, sizeof (wbuf), 50);
		}

	num_packets = 0;

	do {
		int num_bytes_read;
		int num_bytes_written;
		static unsigned char buf[0x2200];

		{
			static char f = 0x0c;
			static char wbuf[] = {0x81, 0x07, 0x87, 0x01, 0x99, 0x0c};
			wbuf[5] = f;
			if (f == 0x0c)
				f = 0x04;
			else
				f = 0x0c;

			num_bytes_written = usb_bulk_write (h, 0x02, wbuf, sizeof (wbuf), 50);
		}

		num_bytes_read = usb_bulk_read (h, 0x81, buf, sizeof (buf), 5000);
		if (num_bytes_read < 0)
			printf ("error doing read: %s (%d)\n", strerror (-num_bytes_read), -num_bytes_read);
		else {
			/*printf ("num_bytes_read = %d\n", num_bytes_read);*/
			hexdump (buf, num_bytes_read);
			printf ("num_packets = %d (bytes_read = %d)\n", num_packets++, num_bytes_read);
			if (buf[0] == 0xe0 || buf[0] == 0x0c0) {
				static int img_num = 0;
				
				dump_image (buf, num_bytes_read, img_num++);
			}
		}

		usleep (100*1000);
	} while (1);

out:
	;
}

int 
main (int argc, char *argv[])
{
    	struct usb_bus *busses;

    	usb_init();
	usb_set_debug(999999);
    	usb_find_busses();
    	usb_find_devices();
    
    	busses = usb_get_busses();


    	struct usb_bus *bus;
    	int c, i, a;
    
    	/* ... */
    
    	for (bus = busses; bus != NULL; bus = bus->next) {
    		struct usb_device *dev;
    
    		for (dev = bus->devices; dev != NULL; dev = dev->next) {
    			if (dev->descriptor.idVendor == 0x08ff &&
			    dev->descriptor.idProduct == 0x5731) {
				printf ("Finger print sensor is at %s\n", dev->filename);
				read_from_sensor (dev);
    			}
    
		}
	}
	
	return 0;
}


#if 0
/*
		num_bytes_written = usb_bulk_write (h, 0x02, wbuf1, sizeof (wbuf1), 5000);
		if (num_bytes_written != 0)
			printf ("error doing write: %s (%d)\n", strerror (-num_bytes_written), -num_bytes_written);
		else
			printf ("num_bytes_written = %d\n", num_bytes_written);

		exit (1);
*/
		num_bytes_written = usb_bulk_write (h, 0x02, wbuf2, sizeof (wbuf2), 50);
/*
		if (num_bytes_written < 0)
			printf ("error doing write: %s (%d)\n", strerror (-num_bytes_written), -num_bytes_written);
		else
			printf ("num_bytes_written = %d\n", num_bytes_written);
*/

		num_bytes_written = usb_bulk_write (h, 0x02, wbuf3, sizeof (wbuf3), 50);
/*		if (num_bytes_written < 0)
			printf ("error doing write: %s (%d)\n", strerror (-num_bytes_written), -num_bytes_written);
		else
			printf ("num_bytes_written = %d\n", num_bytes_written);
*/

		num_bytes_written = usb_bulk_write (h, 0x02, wbuf4, sizeof (wbuf4), 50);
/*		if (num_bytes_written < 0)
			printf ("error doing write: %s (%d)\n", strerror (-num_bytes_written), -num_bytes_written);
		else
			printf ("num_bytes_written = %d\n", num_bytes_written);
*/

		num_bytes_written = usb_bulk_write (h, 0x02, wbuf5, sizeof (wbuf5), 50);
/*		if (num_bytes_written < 0)
			printf ("error doing write: %s (%d)\n", strerror (-num_bytes_written), -num_bytes_written);
		else
			printf ("num_bytes_written = %d\n", num_bytes_written);
*/

		num_bytes_written = usb_bulk_write (h, 0x02, wbuf6, sizeof (wbuf6), 50);
/*		if (num_bytes_written < 0)
			printf ("error doing write: %s (%d)\n", strerror (-num_bytes_written), -num_bytes_written);
		else
			printf ("num_bytes_written = %d\n", num_bytes_written);
*/
#endif
